# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Qiskit backend for the MQT DDSIM deterministic noise-aware simulator."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any, cast

from mqt.core import load
from qiskit.providers import Options
from qiskit.result.models import ExperimentResult, ExperimentResultData

from mqt.ddsim.pyddsim import DeterministicNoiseSimulator

from .experiment_header import DDSIMExperimentHeader
from .qasm_simulator_backend import QasmSimulatorBackend

if TYPE_CHECKING:
    from qiskit import QuantumCircuit


class DeterministicNoiseSimulatorBackend(QasmSimulatorBackend):
    """Qiskit backend for the MQT DDSIM deterministic noise-aware simulator."""

    def __init__(
        self,
        name: str = "deterministic_noise_simulator",
        description: str = "MQT DDSIM deterministic noise-aware simulator",
    ) -> None:
        """Constructor for the MQT DDSIM deterministic noise-aware simulator backend."""
        super().__init__(name=name, description=description)

    @classmethod
    def _default_options(cls) -> Options:
        return Options(
            shots=None,
            parameter_binds=None,
            simulator_seed=None,
            noise_effects="APD",
            noise_probability=0.01,
            amp_damping_probability=0.02,
            multi_qubit_gate_factor=2,
        )

    @staticmethod
    def _run_experiment(qc: QuantumCircuit, **options: dict[str, Any]) -> ExperimentResult:
        start_time = time.time()
        noise_effects = cast("str", options.get("noise_effects", "APD"))
        noise_probability = cast("float", options.get("noise_probability", 0.01))
        amp_damping_probability = cast("float", options.get("amp_damping_probability", 0.02))
        multi_qubit_gate_factor = cast("float", options.get("multi_qubit_gate_factor", 2))
        seed = cast("int", options.get("simulator_seed", -1))
        shots = cast("int", options.get("shots", 1024))

        circ = load(qc)
        sim = DeterministicNoiseSimulator(
            circ=circ,
            seed=seed,
            noise_effects=noise_effects,
            noise_probability=noise_probability,
            amp_damping_probability=amp_damping_probability,
            multi_qubit_gate_factor=multi_qubit_gate_factor,
        )

        counts = sim.simulate(shots=shots)
        end_time = time.time()

        data = ExperimentResultData(
            counts={hex(int(result, 2)): count for result, count in counts.items()},
            statevector=None,
            time_taken=end_time - start_time,
        )

        return ExperimentResult(
            shots=shots,
            success=True,
            status="DONE",
            seed=seed,
            data=data,
            metadata=qc.metadata,
            header=DDSIMExperimentHeader.from_quantum_circuit(qc).get_compatible_version(),
        )



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Qiskit backend for the MQT DDSIM unitary simulator."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

import numpy as np
from mqt.core import load
from qiskit import QiskitError
from qiskit.providers import Options
from qiskit.result.models import ExperimentResult, ExperimentResultData
from qiskit.transpiler import Target

from .experiment_header import DDSIMExperimentHeader
from .pyddsim import UnitarySimulator, UnitarySimulatorMode
from .qasm_simulator_backend import QasmSimulatorBackend
from .target import DDSIMTargetBuilder

if TYPE_CHECKING:
    from collections.abc import Sequence

    from qiskit import QuantumCircuit


class UnitarySimulatorBackend(QasmSimulatorBackend):
    """Qiskit backend for the MQT DDSIM unitary simulator."""

    _US_TARGET = Target(
        description="Target for the MQT DDSIM unitary simulator",
        num_qubits=15,  # corresponds to 16GiB memory for storing the full matrix
    )

    @staticmethod
    def _add_operations_to_target(target: Target) -> None:
        DDSIMTargetBuilder.add_0q_gates(target)
        DDSIMTargetBuilder.add_1q_gates(target)
        DDSIMTargetBuilder.add_2q_gates(target)
        DDSIMTargetBuilder.add_3q_gates(target)
        DDSIMTargetBuilder.add_multi_qubit_gates(target)
        DDSIMTargetBuilder.add_barrier(target)

    def __init__(
        self,
        name: str = "unitary_simulator",
        description: str = "MQT DDSIM unitary simulator",
    ) -> None:
        """Constructor for the MQT DDSIM unitary simulator backend."""
        super().__init__(name=name, description=description)

    @classmethod
    def _default_options(cls) -> Options:
        return Options(shots=1, mode="recursive", parameter_binds=None)

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._US_TARGET

    @classmethod
    def _run_experiment(cls, qc: QuantumCircuit, **options: Any) -> ExperimentResult:
        start_time = time.time()
        seed = options.get("seed", -1)
        mode = options.get("mode", "recursive")

        if mode == "sequential":
            construction_mode = UnitarySimulatorMode.sequential
        elif mode == "recursive":
            construction_mode = UnitarySimulatorMode.recursive
        else:
            msg = (
                f"Construction mode {mode} not supported by DDSIM unitary simulator. Available modes are "
                "'recursive' and 'sequential'"
            )
            raise QiskitError(msg)

        circuit = load(qc)
        sim = UnitarySimulator(circuit, seed=seed, mode=construction_mode)
        sim.construct()
        # Extract resulting matrix from final DD and write data
        dd = sim.get_constructed_dd()
        mat = dd.get_matrix(sim.get_number_of_qubits())
        unitary = np.array(mat, copy=False)
        end_time = time.time()

        data = ExperimentResultData(
            unitary=unitary,
            construction_time=sim.get_construction_time(),
            dd_nodes=sim.get_final_node_count(),
            time_taken=end_time - start_time,
        )

        return ExperimentResult(
            shots=1,
            success=True,
            status="DONE",
            seed=seed,
            data=data,
            metadata=qc.metadata,
            header=DDSIMExperimentHeader.from_quantum_circuit(qc).get_compatible_version(),
        )

    def _validate(self, quantum_circuits: Sequence[QuantumCircuit]) -> None:
        """Semantic validations of the quantum circuits which cannot be done via schemas.

        1. No shots
        2. No measurements in the middle.
        """
        for qc in quantum_circuits:
            name = qc.name
            n_qubits = qc.num_qubits
            max_qubits = self.target.num_qubits

            if n_qubits > max_qubits:
                msg = f"Number of qubits {n_qubits} is greater than maximum ({max_qubits}) for '{self.name}'."
                raise QiskitError(msg)

            if qc.metadata is not None and "shots" in qc.metadata and qc.metadata["shots"] != 1:
                qc.metadata["shots"] = 1

            for instruction in qc.data:
                operation = instruction.operation
                if operation.name in {"measure", "reset"}:
                    msg = f"Unsupported '{self.name}' instruction '{operation.name}' in circuit '{name}'."
                    raise QiskitError(msg)



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Qiskit backend for the MQT DDSIM simulation-path QASM simulator."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from qiskit import QuantumCircuit


import numpy as np
from mqt.core import load
from qiskit.providers import Options
from qiskit.result.models import ExperimentResult, ExperimentResultData
from qiskit.transpiler import Target

from .experiment_header import DDSIMExperimentHeader
from .pyddsim import PathSimulator, PathSimulatorConfiguration
from .qasm_simulator_backend import QasmSimulatorBackend
from .target import DDSIMTargetBuilder


class PathQasmSimulatorBackend(QasmSimulatorBackend):
    """Qiskit backend for the MQT DDSIM simulation-path QASM simulator."""

    _PATH_TARGET = Target(
        description="Target for the MQT DDSIM simulation-path QASM simulator",
        num_qubits=128,
    )

    @staticmethod
    def _add_operations_to_target(target: Target) -> None:
        DDSIMTargetBuilder.add_0q_gates(target)
        DDSIMTargetBuilder.add_1q_gates(target)
        DDSIMTargetBuilder.add_2q_gates(target)
        DDSIMTargetBuilder.add_3q_gates(target)
        DDSIMTargetBuilder.add_multi_qubit_gates(target)
        DDSIMTargetBuilder.add_barrier(target)
        DDSIMTargetBuilder.add_measure(target)

    def __init__(
        self,
        name: str = "path_qasm_simulator",
        description: str = "MQT DDSIM simulation-path QASM simulator",
    ) -> None:
        """Constructor for the MQT DDSIM simulation-path QASM simulator backend."""
        super().__init__(name=name, description=description)

    @classmethod
    def _default_options(cls) -> Options:
        return Options(
            shots=None,
            parameter_binds=None,
            simulator_seed=None,
            pathsim_configuration=PathSimulatorConfiguration(),
            mode=None,
            bracket_size=None,
            alternating_start=None,
            gate_cost=None,
            seed=None,
        )

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._PATH_TARGET

    def _run_experiment(self, qc: QuantumCircuit, **options: Any) -> ExperimentResult:
        start_time = time.time()

        pathsim_configuration = options.get("pathsim_configuration", PathSimulatorConfiguration())

        mode = options.get("mode")
        if mode is not None:
            pathsim_configuration.mode = mode

        bracket_size = options.get("bracket_size")
        if bracket_size is not None:
            pathsim_configuration.bracket_size = bracket_size

        alternating_start = options.get("alternating_start")
        if alternating_start is not None:
            pathsim_configuration.alternating_start = alternating_start

        gate_cost = options.get("gate_cost")
        if gate_cost is not None:
            pathsim_configuration.gate_cost = gate_cost

        seed: int | None = options.get("seed")
        if seed is not None:
            pathsim_configuration.seed = seed

        circuit = load(qc)
        sim = PathSimulator(circuit, config=pathsim_configuration)

        shots = options.get("shots", 1024)
        setup_time = time.time()
        counts = sim.simulate(shots)
        end_time = time.time()

        data = ExperimentResultData(
            counts={hex(int(result, 2)): count for result, count in counts.items()},
            statevector=None if not self._SHOW_STATE_VECTOR else np.array(sim.get_constructed_dd().get_vector()),
            time_taken=end_time - start_time,
            time_setup=setup_time - start_time,
            time_sim=end_time - setup_time,
        )

        return ExperimentResult(
            shots=shots,
            success=True,
            status="DONE",
            config=pathsim_configuration,
            seed=seed,
            data=data,
            metadata=qc.metadata,
            header=DDSIMExperimentHeader.from_quantum_circuit(qc).get_compatible_version(),
        )



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Module for Qiskit Primitives."""

from __future__ import annotations

from .estimator import Estimator
from .sampler import Sampler

__all__ = ["Estimator", "Sampler"]



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Sampler implementation using QasmSimulatorBackend."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from mqt.core import load
from qiskit.circuit import QuantumCircuit
from qiskit.primitives.base import BaseSamplerV2
from qiskit.primitives.containers import (
    BitArray,
    DataBin,
    PrimitiveResult,
    SamplerPubResult,
)
from qiskit.primitives.containers.sampler_pub import SamplerPub
from qiskit.primitives.primitive_job import PrimitiveJob

from mqt.ddsim.pyddsim import CircuitSimulator

if TYPE_CHECKING:
    from collections.abc import Iterable

    from qiskit.circuit import ClassicalRegister
    from qiskit.primitives.containers import SamplerPubLike


class Sampler(BaseSamplerV2):  # type: ignore[misc]
    """DDSIM implementation of Qiskit's sampler.

    The implementation is adapted from Qiskit's `StatevectorSampler`.
    """

    def __init__(self, *, default_shots: int = 1024, seed: int = -1) -> None:
        """Create a new DDSIM sampler.

        Args:
            default_shots: The default number of shots to sample. Defaults to ``1024``.
            seed: The seed for the ``CircuitSimulator``. Defaults to ``-1``.
        """
        self._default_shots = default_shots
        self._seed = seed

    @property
    def default_shots(self) -> int:
        """Return the default shots."""
        return self._default_shots

    @property
    def seed(self) -> int:
        """Return the seed."""
        return self._seed

    def run(
        self,
        pubs: Iterable[SamplerPubLike],
        *,
        shots: int | None = None,
    ) -> PrimitiveJob[PrimitiveResult[SamplerPubResult]]:
        """Run and collect samples from each provided PUB (primitive unified bloc).

        Each PUB is run on the ``CircuitSimulator``.

        Args:
            pubs: An iterable of pub-like objects, such as tuples ``(circuit, parameter_values)``.
            shots: The number of shots to sample. If ``None``, the default number of shots is used.
        """
        if shots is None:
            shots = self._default_shots
        coerced_pubs = [SamplerPub.coerce(pub, shots) for pub in pubs]

        job = PrimitiveJob(self._run, coerced_pubs)
        job._submit()  # noqa: SLF001
        return job

    def _run(self, pubs: Iterable[SamplerPub]) -> PrimitiveResult[SamplerPubResult]:
        results = [self._run_pub(pub) for pub in pubs]
        return PrimitiveResult(results, metadata={"version": 2})

    def _run_pub(self, pub: SamplerPub) -> SamplerPubResult:
        circuit = pub.circuit
        parameter_values = pub.parameter_values
        shots = pub.shots

        bound_circuits = parameter_values.bind_all(circuit)
        bound_circuits_list = np.asarray(bound_circuits, dtype=object).tolist()
        if isinstance(bound_circuits_list, QuantumCircuit):
            bound_circuits_list = [bound_circuits_list]

        counts = self._run_experiment(bound_circuits_list, shots=shots, seed=self.seed)

        bit_arrays = self._get_bit_arrays(circuit.cregs, counts)
        return SamplerPubResult(
            DataBin(**bit_arrays, shape=pub.shape),
            metadata={"shots": shots, "circuit_metadata": circuit.metadata},
        )

    @staticmethod
    def _run_experiment(quantum_circuits: list[QuantumCircuit], shots: int, seed: int) -> list[dict[str, int]]:
        counts_list = []

        for quantum_circuit in quantum_circuits:
            loaded_quantum_circuit = load(quantum_circuit)
            simulator = CircuitSimulator(loaded_quantum_circuit, seed=seed)
            counts = simulator.simulate(shots)
            counts_list.append(counts)

        return counts_list

    @staticmethod
    def _get_bit_arrays(
        cregs: list[ClassicalRegister],
        counts: list[dict[str, int]],
    ) -> dict[str, BitArray]:
        bit_arrays = {}

        start_index = 0
        for creg in cregs:
            creg_counts = [
                {
                    key[start_index - creg.size : start_index if start_index != 0 else None]: value
                    for key, value in count.items()
                }
                for count in counts
            ]
            bit_arrays[creg.name] = BitArray.from_counts(creg_counts, creg.size)
            start_index -= creg.size

        return bit_arrays



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Qiskit backend for the MQT DDSIM hybrid Schrodinger-Feynman statevector simulator."""

from __future__ import annotations

from qiskit.transpiler import Target

from .hybrid_qam_simulator_backend import HybridQasmSimulatorBackend


class HybridStatevectorSimulatorBackend(HybridQasmSimulatorBackend):
    """Qiskit backend for the MQT DDSIM hybrid Schrodinger-Feynman statevector simulator."""

    _SHOW_STATE_VECTOR = True
    _HSF_SV_TARGET = Target(
        description="Target for the MQT DDSIM hybrid Schrodinger-Feynman statevector simulator",
        num_qubits=30,  # corresponds to 16GiB memory for storing the full statevector
    )

    def __init__(
        self,
        name: str = "hybrid_statevector_simulator",
        description: str = "MQT DDSIM hybrid Schrodinger-Feynman statevector simulator",
    ) -> None:
        """Constructor for the MQT DDSIM hybrid Schrodinger-Feynman statevector simulator backend."""
        super().__init__(
            name=name,
            description=description,
        )

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._HSF_SV_TARGET



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Qiskit backend for the MQT DDSIM hybrid Schrodinger-Feynman simulator."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from qiskit import QuantumCircuit

import numpy as np
from mqt.core import load
from qiskit import QiskitError
from qiskit.providers import Options
from qiskit.result.models import ExperimentResult, ExperimentResultData
from qiskit.transpiler import Target
from qiskit.utils import local_hardware_info

from .experiment_header import DDSIMExperimentHeader
from .pyddsim import HybridSimulator, HybridSimulatorMode
from .qasm_simulator_backend import QasmSimulatorBackend
from .target import DDSIMTargetBuilder


class HybridQasmSimulatorBackend(QasmSimulatorBackend):
    """Qiskit backend for the MQT DDSIM hybrid Schrodinger-Feynman simulator."""

    _HSF_TARGET = Target(
        description="Target for the MQT DDSIM Schrodinger-Feynman simulator",
        num_qubits=128,
    )

    @staticmethod
    def _add_operations_to_target(target: Target) -> None:
        DDSIMTargetBuilder.add_0q_gates(target)
        DDSIMTargetBuilder.add_1q_gates(target)
        DDSIMTargetBuilder.add_2q_controlled_gates(target)
        DDSIMTargetBuilder.add_barrier(target)
        DDSIMTargetBuilder.add_measure(target)

    def __init__(
        self,
        name: str = "hybrid_qasm_simulator",
        description: str = "MQT DDSIM hybrid Schrodinger-Feynman QASM simulator",
    ) -> None:
        """Constructor for the MQT DDSIM hybrid Schrodinger-Feynman QASM simulator backend."""
        super().__init__(name=name, description=description)

    @classmethod
    def _default_options(cls) -> Options:
        return Options(
            shots=None,
            parameter_binds=None,
            simulator_seed=None,
            mode="amplitude",
            nthreads=local_hardware_info()["cpus"],
        )

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._HSF_TARGET

    def _run_experiment(self, qc: QuantumCircuit, **options: Any) -> ExperimentResult:
        start_time = time.time()
        seed = options.get("seed", -1)
        mode = options.get("mode", "amplitude")
        nthreads = int(options.get("nthreads", local_hardware_info()["cpus"]))

        if mode == "amplitude":
            hybrid_mode = HybridSimulatorMode.amplitude
            max_qubits = 30  # hard-coded 16GiB memory limit
            algorithm_qubits = qc.num_qubits
            if algorithm_qubits > max_qubits:
                msg = "Not enough memory available to simulate the circuit even on a single thread"
                raise QiskitError(msg)
            qubit_diff = max_qubits - algorithm_qubits
            nthreads = int(min(2**qubit_diff, nthreads))
        elif mode == "dd":
            hybrid_mode = HybridSimulatorMode.DD
        else:
            msg = f"Simulation mode{mode} not supported by hybrid simulator. Available modes are 'amplitude' and 'dd'."
            raise QiskitError(msg)

        circuit = load(qc)
        sim = HybridSimulator(circuit, seed=seed, mode=hybrid_mode, nthreads=nthreads)

        shots = options.get("shots", 1024)
        if self._SHOW_STATE_VECTOR and shots > 0:
            shots = 0

        counts = sim.simulate(shots)
        end_time = time.time()

        data = ExperimentResultData(
            counts={hex(int(result, 2)): count for result, count in counts.items()},
            statevector=None
            if not self._SHOW_STATE_VECTOR
            else np.array(sim.get_constructed_dd().get_vector(), copy=False)
            if sim.get_mode() == HybridSimulatorMode.DD
            else sim.get_final_amplitudes(),
            time_taken=end_time - start_time,
            mode=mode,
            nthreads=nthreads,
        )

        return ExperimentResult(
            shots=shots,
            success=True,
            status="DONE",
            seed=seed,
            data=data,
            metadata=qc.metadata,
            header=DDSIMExperimentHeader.from_quantum_circuit(qc).get_compatible_version(),
        )



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Site customization shim to enable multiprocess coverage collection in tests.

See https://coverage.readthedocs.io/en/latest/subprocess.html.
"""

from __future__ import annotations

try:
    import coverage

    coverage.process_startup()
except ImportError:
    # The 'coverage' module is optional
    # If it is not installed, we do not enable multiprocess coverage collection
    pass



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Qiskit backend for MQT DDSIM statevector simulators."""

from __future__ import annotations

from qiskit.transpiler import Target

from .qasm_simulator_backend import QasmSimulatorBackend


class StatevectorSimulatorBackend(QasmSimulatorBackend):
    """Qiskit backend for MQT DDSIM statevector simulators."""

    _SHOW_STATE_VECTOR = True
    _SV_TARGET = Target(
        description="Target for the MQT DDSIM statevector simulator",
        num_qubits=30,  # corresponds to 16GiB memory for storing the full statevector
    )

    def __init__(
        self,
        name: str = "statevector_simulator",
        description: str = "MQT DDSIM statevector simulator",
    ) -> None:
        """Constructor for the MQT DDSIM statevector simulator backend."""
        super().__init__(name=name, description=description)

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._SV_TARGET



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

from enum import Enum
from typing import Any, overload

from mqt.core.dd import MatrixDD, VectorDD
from mqt.core.ir import QuantumComputation

__all__ = [
    "CircuitSimulator",
    "DeterministicNoiseSimulator",
    "HybridSimulator",
    "HybridSimulatorMode",
    "PathSimulator",
    "PathSimulatorConfiguration",
    "PathSimulatorMode",
    "StochasticNoiseSimulator",
    "UnitarySimulator",
    "UnitarySimulatorMode",
]

class CircuitSimulator:
    def __init__(
        self,
        circ: QuantumComputation,
        approximation_step_fidelity: float = 1.0,
        approximation_steps: int = 1,
        approximation_strategy: str = "fidelity",
        seed: int = -1,
    ) -> None: ...
    def expectation_value(self, observable: QuantumComputation) -> float: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def get_constructed_dd(self) -> VectorDD: ...
    def set_tolerance(self, tol: float) -> None: ...
    def simulate(self, shots: int) -> dict[str, int]: ...
    def statistics(self) -> dict[str, str]: ...

class StochasticNoiseSimulator:
    def __init__(
        self,
        circ: QuantumComputation,
        approximation_step_fidelity: float = 1.0,
        approximation_steps: int = 1,
        approximation_strategy: str = "fidelity",
        seed: int = -1,
        noise_effects: str = "APD",
        noise_probability: float = 0.01,
        amp_damping_probability: float | None = 0.02,
        multi_qubit_gate_factor: float = 2,
    ) -> None: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def get_constructed_dd(self) -> VectorDD: ...
    def set_tolerance(self, tol: float) -> None: ...
    def simulate(self, shots: int) -> dict[str, int]: ...
    def statistics(self) -> dict[str, str]: ...

class DeterministicNoiseSimulator:
    def __init__(
        self,
        circ: QuantumComputation,
        approximation_step_fidelity: float = 1.0,
        approximation_steps: int = 1,
        approximation_strategy: str = "fidelity",
        seed: int = -1,
        noise_effects: str = "APD",
        noise_probability: float = 0.01,
        amp_damping_probability: float | None = 0.02,
        multi_qubit_gate_factor: float = 2,
    ) -> None: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def get_constructed_dd(self) -> VectorDD: ...
    def set_tolerance(self, tol: float) -> None: ...
    def simulate(self, shots: int) -> dict[str, int]: ...
    def statistics(self) -> dict[str, str]: ...

class HybridSimulatorMode(Enum):
    """Enumeration of modes for the :class:`~HybridSimulator`."""

    DD = ...
    amplitude = ...

class HybridSimulator:
    def __init__(
        self,
        circ: QuantumComputation,
        approximation_step_fidelity: float = 1.0,
        approximation_steps: int = 1,
        approximation_strategy: str = "fidelity",
        seed: int = -1,
        mode: HybridSimulatorMode = ...,
        nthreads: int = 2,
    ) -> None: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_final_amplitudes(self) -> list[complex]: ...
    def get_mode(self) -> HybridSimulatorMode: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def get_constructed_dd(self) -> VectorDD: ...
    def set_tolerance(self, tol: float) -> None: ...
    def simulate(self, shots: int) -> dict[str, int]: ...
    def statistics(self) -> dict[str, str]: ...

class PathSimulatorMode(Enum):
    """Enumeration of modes for the :class:`~PathCSimulator`."""

    alternating = ...
    bracket = ...
    gate_cost = ...
    pairwise_recursive = ...
    sequential = ...

class PathSimulatorConfiguration:
    def __init__(self) -> None: ...
    def json(self) -> dict[str, Any]: ...
    @property
    def bracket_size(self) -> int: ...
    @bracket_size.setter
    def bracket_size(self, arg0: int) -> None: ...
    @property
    def gate_cost(self) -> list[int]: ...
    @gate_cost.setter
    def gate_cost(self, arg0: list[int]) -> None: ...
    @property
    def mode(self) -> PathSimulatorMode: ...
    @mode.setter
    def mode(self, arg0: PathSimulatorMode) -> None: ...
    @property
    def seed(self) -> int: ...
    @seed.setter
    def seed(self, arg0: int) -> None: ...
    @property
    def starting_point(self) -> int: ...
    @starting_point.setter
    def starting_point(self, arg0: int) -> None: ...

class PathSimulator:
    @overload
    def __init__(self, circ: QuantumComputation, config: PathSimulatorConfiguration = ...) -> None: ...
    @overload
    def __init__(
        self,
        circ: QuantumComputation,
        mode: PathSimulatorMode = ...,
        bracket_size: int = 2,
        starting_point: int = 0,
        gate_cost: list[int] = ...,
        seed: int = ...,
    ) -> None: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def get_constructed_dd(self) -> VectorDD: ...
    def set_simulation_path(self, path: list[tuple[int, int]], assume_correct_order: bool = False) -> None: ...
    def set_tolerance(self, tol: float) -> None: ...
    def simulate(self, shots: int) -> dict[str, int]: ...
    def statistics(self) -> dict[str, str]: ...

class UnitarySimulatorMode(Enum):
    """Enumeration of modes for the :class:`~UnitarySimulator`."""

    recursive = ...
    sequential = ...

class UnitarySimulator:
    def __init__(
        self,
        circ: QuantumComputation,
        approximation_step_fidelity: float = 1.0,
        approximation_steps: int = 1,
        approximation_strategy: str = "fidelity",
        seed: int = -1,
        mode: UnitarySimulatorMode = ...,
    ) -> None: ...
    def construct(self) -> None: ...
    def get_active_matrix_node_count(self) -> int: ...
    def get_active_vector_node_count(self) -> int: ...
    def get_construction_time(self) -> float: ...
    def get_final_node_count(self) -> int: ...
    def get_max_node_count(self) -> int: ...
    def get_mode(self) -> UnitarySimulatorMode: ...
    def get_name(self) -> str: ...
    def get_number_of_qubits(self) -> int: ...
    def get_tolerance(self) -> float: ...
    def set_tolerance(self, tol: float) -> None: ...
    def statistics(self) -> dict[str, str]: ...
    def get_constructed_dd(self) -> MatrixDD: ...



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Utilities for constructing a DDSIM target."""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

import qiskit.circuit.library as qcl
from qiskit.circuit import Parameter

if TYPE_CHECKING:
    from qiskit.transpiler import Target


class DDSIMTargetBuilder:
    """Class for building DDSIM targets."""

    @classmethod
    def add_0q_gates(cls, target: Target) -> None:
        """Add 0-qubit gates to the target."""
        target.add_instruction(qcl.GlobalPhaseGate(Parameter("phase")))

    @classmethod
    def add_1q_clifford_gates(cls, target: Target) -> None:
        """Add 1-qubit Clifford gates to the target."""
        target.add_instruction(qcl.IGate())
        target.add_instruction(qcl.XGate())
        target.add_instruction(qcl.YGate())
        target.add_instruction(qcl.ZGate())
        target.add_instruction(qcl.HGate())
        target.add_instruction(qcl.SGate())
        target.add_instruction(qcl.SdgGate())
        target.add_instruction(qcl.SXGate())
        target.add_instruction(qcl.SXdgGate())

    @classmethod
    def add_1q_gates(cls, target: Target) -> None:
        """Add 1-qubit gates to the target."""
        cls.add_1q_clifford_gates(target)

        theta = Parameter("theta")
        phi = Parameter("phi")
        lam = Parameter("lam")

        target.add_instruction(qcl.TGate())
        target.add_instruction(qcl.TdgGate())
        target.add_instruction(qcl.RXGate(theta))
        target.add_instruction(qcl.RYGate(theta))
        target.add_instruction(qcl.RZGate(phi))
        target.add_instruction(qcl.PhaseGate(theta))
        target.add_instruction(qcl.U1Gate(theta))
        target.add_instruction(qcl.U2Gate(phi, lam))
        target.add_instruction(qcl.U3Gate(theta, phi, lam))

    @classmethod
    def add_2q_controlled_clifford_gates(cls, target: Target) -> None:
        """Add 2-qubit controlled Clifford gates to the target."""
        target.add_instruction(qcl.CXGate())
        target.add_instruction(qcl.CYGate())
        target.add_instruction(qcl.CZGate())

    @classmethod
    def add_2q_controlled_gates(cls, target: Target) -> None:
        """Add 2-qubit controlled gates to the target."""
        cls.add_2q_controlled_clifford_gates(target)

        theta = Parameter("theta")
        phi = Parameter("phi")
        lam = Parameter("lam")

        target.add_instruction(qcl.CHGate())
        target.add_instruction(qcl.CSXGate())
        target.add_instruction(qcl.CRXGate(theta))
        target.add_instruction(qcl.CRYGate(theta))
        target.add_instruction(qcl.CRZGate(theta))
        target.add_instruction(qcl.CPhaseGate(theta))
        target.add_instruction(qcl.CU1Gate(theta))
        target.add_instruction(qcl.CU3Gate(theta, phi, lam))
        with contextlib.suppress(AttributeError):
            target.add_instruction(qcl.CSGate())
            target.add_instruction(qcl.CSdgGate())

    @classmethod
    def add_2q_non_controlled_clifford_gates(cls, target: Target) -> None:
        """Add 2-qubit non-controlled Clifford gates to the target."""
        target.add_instruction(qcl.SwapGate())
        target.add_instruction(qcl.iSwapGate())
        target.add_instruction(qcl.DCXGate())
        target.add_instruction(qcl.ECRGate())

    @classmethod
    def add_2q_non_controlled_gates(cls, target: Target) -> None:
        """Add 2-qubit non-controlled gates to the target."""
        cls.add_2q_non_controlled_clifford_gates(target)

        beta = Parameter("beta")
        theta = Parameter("theta")

        target.add_instruction(qcl.RXXGate(theta))
        target.add_instruction(qcl.RYYGate(theta))
        target.add_instruction(qcl.RZZGate(theta))
        target.add_instruction(qcl.RZXGate(theta))
        target.add_instruction(qcl.XXMinusYYGate(theta, beta))
        target.add_instruction(qcl.XXPlusYYGate(theta, beta))

    @classmethod
    def add_2q_gates(cls, target: Target) -> None:
        """Add 2-qubit gates to the target."""
        cls.add_2q_controlled_gates(target)
        cls.add_2q_non_controlled_gates(target)

    @classmethod
    def add_3q_gates(cls, target: Target) -> None:
        """Add 3-qubit gates to the target."""
        target.add_instruction(qcl.CCXGate())
        target.add_instruction(qcl.CSwapGate())
        with contextlib.suppress(AttributeError):
            target.add_instruction(qcl.CCZGate())

    @classmethod
    def add_multi_qubit_gates(cls, target: Target) -> None:
        """Add multi-qubit gates to the target."""
        target.add_instruction(qcl.MCXGate, name="mcx")
        target.add_instruction(qcl.MCXGrayCode, name="mcx_gray")
        target.add_instruction(qcl.MCXRecursive, name="mcx_recursive")
        target.add_instruction(qcl.MCXVChain, name="mcx_vchain")
        target.add_instruction(qcl.MCPhaseGate, name="mcphase")
        target.add_instruction(qcl.MCU1Gate, name="mcu1")

    @classmethod
    def add_measure(cls, target: Target) -> None:
        """Add a measure instruction to the target."""
        target.add_instruction(qcl.Measure())

    @classmethod
    def add_reset(cls, target: Target) -> None:
        """Add a reset instruction to the target."""
        target.add_instruction(qcl.Reset())

    @classmethod
    def add_non_unitary_operations(cls, target: Target) -> None:
        """Add non-unitary operations to the target."""
        cls.add_measure(target)
        cls.add_reset(target)

    @classmethod
    def add_barrier(cls, target: Target) -> None:
        """Add a barrier instruction to the target."""
        target.add_instruction(qcl.Barrier, name="barrier")



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Provider for DDSIM backends."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from qiskit.providers.exceptions import QiskitBackendNotFoundError
from qiskit.providers.providerutils import filter_backends

from .deterministic_noise_simulator_backend import DeterministicNoiseSimulatorBackend
from .hybrid_qam_simulator_backend import HybridQasmSimulatorBackend
from .hybrid_statevector_simulator_backend import HybridStatevectorSimulatorBackend
from .path_qasm_simulator_backend import PathQasmSimulatorBackend
from .path_statevector_simulator_backend import PathStatevectorSimulatorBackend
from .qasm_simulator_backend import QasmSimulatorBackend
from .statevector_simulator_backend import StatevectorSimulatorBackend
from .stochastic_noise_simulator_backend import StochasticNoiseSimulatorBackend
from .unitary_simulator_backend import UnitarySimulatorBackend

if TYPE_CHECKING:
    from collections.abc import Callable

    from qiskit.providers import BackendV2


class DDSIMProvider:
    """Provider for DDSIM backends."""

    _BACKENDS = (
        ("qasm_simulator", QasmSimulatorBackend),
        ("statevector_simulator", StatevectorSimulatorBackend),
        ("hybrid_qasm_simulator", HybridQasmSimulatorBackend),
        ("hybrid_statevector_simulator", HybridStatevectorSimulatorBackend),
        ("path_qasm_simulator", PathQasmSimulatorBackend),
        ("path_statevector_simulator", PathStatevectorSimulatorBackend),
        ("unitary_simulator", UnitarySimulatorBackend),
        ("stochastic_noise_simulator", StochasticNoiseSimulatorBackend),
        ("deterministic_noise_simulator", DeterministicNoiseSimulatorBackend),
    )

    def get_backend(self, name: str | None = None, **kwargs: Any) -> BackendV2:
        """Return a backend matching the specified criteria.

        Args:
            name: Name of the backend.
            kwargs: Additional filtering criteria.
        """
        backends = self.backends(name, **kwargs)
        if len(backends) > 1:
            msg = "More than one backend matches the criteria"
            raise QiskitBackendNotFoundError(msg)
        if not backends:
            msg = "No backend matches the criteria"
            raise QiskitBackendNotFoundError(msg)

        return backends[0]

    def backends(
        self,
        name: str | None = None,
        filters: Callable[[list[BackendV2]], list[BackendV2]] | None = None,
        **kwargs: dict[str, Any],
    ) -> list[BackendV2]:
        """Return a list of backends matching the specified criteria.

        Args:
            name: Name of the backend.
            filters: Additional filtering criteria.
            kwargs: Additional filtering criteria.
        """
        backends = [
            backend_cls() for backend_name, backend_cls in self._BACKENDS if name is None or backend_name == name
        ]
        return cast("list[BackendV2]", filter_backends(backends, filters=filters, **kwargs))

    def __str__(self) -> str:
        """Return the provider name."""
        return "DDSIMProvider"



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Estimator implementation using DDSIM CircuitSimulator."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from mqt.core import load
from qiskit.circuit import QuantumCircuit
from qiskit.primitives.base import BaseEstimatorV2
from qiskit.primitives.containers import DataBin, PrimitiveResult, PubResult
from qiskit.primitives.containers.estimator_pub import EstimatorPub
from qiskit.primitives.primitive_job import PrimitiveJob
from qiskit.quantum_info import Pauli, SparsePauliOp

from mqt.ddsim.pyddsim import CircuitSimulator

if TYPE_CHECKING:
    from collections.abc import Iterable

    from numpy.typing import NDArray
    from qiskit.primitives.container import ObservablesArray
    from qiskit.primitives.containers import EstimatorPubLike
    from qiskit.quantum_info import Pauli


class Estimator(BaseEstimatorV2):  # type: ignore[misc]
    """DDSIM implementation of Qiskit's estimator.

    The implementation is adapted from Qiskit's `StatevectorEstimator`.
    """

    def __init__(self, *, default_precision: float = 0.0, seed: int = -1) -> None:
        """Create a new DDSIM estimator.

        Args:
            default_precision: The default precision for expectation-value estimates. Defaults to ``0.0``.
            seed: The seed for the ``CircuitSimulator``. Defaults to ``-1``.
        """
        self._default_precision = default_precision
        self._seed = seed

    @property
    def default_precision(self) -> float:
        """Return the default precision."""
        return self._default_precision

    @property
    def seed(self) -> int:
        """Return the seed."""
        return self._seed

    def run(
        self,
        pubs: Iterable[EstimatorPubLike],
        *,
        precision: float | None = None,
    ) -> PrimitiveJob[PrimitiveResult[PubResult]]:
        """Estimate expectation values for each provided PUB (primitive unified bloc).

        Each PUB is run on the CircuitSimulator.

        Args:
            pubs: An iterable of pub-like objects, such as tuples ``(circuit, observables)`` or ``(circuit, observables, parameter_values)``.
            precision: The target precision for expectation-value estimates. If ``None``, the default precision is used.
        """
        if precision is None:
            precision = self._default_precision
        coerced_pubs = [EstimatorPub.coerce(pub, precision) for pub in pubs]

        job = PrimitiveJob(self._run, coerced_pubs)
        job._submit()  # noqa: SLF001
        return job

    def _run(self, pubs: list[EstimatorPub]) -> PrimitiveResult[PubResult]:
        return PrimitiveResult([self._run_pub(pub) for pub in pubs], metadata={"version": 2})

    def _get_observable_circuits(
        self,
        observables: ObservablesArray,
        num_qubits: int,
    ) -> NDArray[np.object_]:
        """Get the quantum-circuit representations of the obvervables."""
        observable_circuits = np.zeros_like(observables, dtype=object)

        for index in np.ndindex(*observables.shape):
            observable = observables[index]

            pauli_strings, coeffs = zip(*observable.items(), strict=False)
            paulis = SparsePauliOp(pauli_strings, coeffs).paulis

            observable_circuits_list = []
            for pauli in paulis:
                qubit_indices = self._get_qubit_indices(pauli)
                observable_circuit = self._get_observable_circuit(pauli, num_qubits, qubit_indices)
                observable_circuits_list.append(observable_circuit)

            observable_circuits[index] = (coeffs, observable_circuits_list)

        return observable_circuits

    @staticmethod
    def _get_qubit_indices(pauli: Pauli) -> list[int]:
        """Get the indices of the qubits that are part of the Pauli observable."""
        qubit_indices = np.arange(pauli.num_qubits)[pauli.z | pauli.x]

        if not np.any(qubit_indices):
            return [0]

        qubit_indices_list = qubit_indices.tolist()
        if isinstance(qubit_indices_list, int):
            qubit_indices_list = [qubit_indices_list]

        return qubit_indices_list  # type: ignore[no-any-return]

    @staticmethod
    def _get_observable_circuit(pauli: Pauli, num_qubits: int, qubit_indices: list[int]) -> QuantumCircuit:
        """Get the quantum-circuit representation of a Pauli observable."""
        observable_circuit = QuantumCircuit(num_qubits, len(qubit_indices))
        for i in qubit_indices:
            if pauli.x[i]:
                if pauli.z[i]:
                    observable_circuit.y(i)
                else:
                    observable_circuit.x(i)
            elif pauli.z[i]:
                observable_circuit.z(i)

        return observable_circuit

    def _run_pub(self, pub: EstimatorPub) -> PubResult:
        circuit = pub.circuit
        observables = pub.observables
        parameter_values = pub.parameter_values

        observable_circuits = self._get_observable_circuits(observables, circuit.num_qubits)
        bound_circuits = parameter_values.bind_all(circuit)
        bc_bound_circuits, bc_observable_circuits = np.broadcast_arrays(bound_circuits, observable_circuits)

        evs = np.zeros_like(bc_bound_circuits, dtype=np.float64)
        stds = np.zeros_like(bc_bound_circuits, dtype=np.float64)

        for index in np.ndindex(*bc_bound_circuits.shape):
            bound_circuit = bc_bound_circuits[index]
            observable_coeffs, observable_circuits = bc_observable_circuits[index]
            expectation_values = self._run_experiment(bound_circuit, observable_circuits, self.seed)
            evs[index] = np.dot(expectation_values, observable_coeffs)

        data = DataBin(evs=evs, stds=stds, shape=evs.shape)
        return PubResult(data)

    @staticmethod
    def _run_experiment(
        bound_circuit: QuantumCircuit,
        observable_circuits: list[QuantumCircuit],
        seed: int = -1,
    ) -> list[float]:
        qc = load(bound_circuit)
        sim = CircuitSimulator(qc, seed=seed)

        return [
            sim.expectation_value(observable=load(observable_circuit)) for observable_circuit in observable_circuits
        ]



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Job module for DDSIM backend."""

from __future__ import annotations

import functools
from concurrent import futures
from typing import TYPE_CHECKING, Any

from qiskit.providers import JobError, JobStatus, JobV1

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence

    from qiskit import QuantumCircuit
    from qiskit.circuit import Parameter
    from qiskit.circuit.parameterexpression import ParameterValueType
    from qiskit.providers import BackendV2
    from qiskit.result import Result

    Parameters = Mapping[Parameter, ParameterValueType] | Sequence[ParameterValueType]


def requires_submit(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to ensure that a submit has been performed before calling the method.

    Args:
        func (callable): test function to be decorated.

    Returns:
        callable: the decorated function.
    """

    @functools.wraps(func)
    def _wrapper(self: DDSIMJob, *args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        if self._future is None:
            msg = "Job not submitted yet!. You have to .submit() first!"
            raise JobError(msg)
        return func(self, *args, **kwargs)

    return _wrapper


class DDSIMJob(JobV1):  # type: ignore[misc]
    """DDSIMJob class.

    Attributes:
        _executor (futures.Executor): executor to handle asynchronous jobs
    """

    _executor = futures.ThreadPoolExecutor(max_workers=1)

    def __init__(
        self,
        backend: BackendV2,
        job_id: str,
        fn: Callable[..., Result],
        experiments: Sequence[QuantumCircuit],
        parameter_values: Sequence[Parameters] | None,
        **args: dict[str, Any],
    ) -> None:
        """Constructor.

        Args:
            backend: the backend instance used to run the job.
            job_id: job ID.
            fn: function to be run by the job.
            experiments: circuits to be executed.
            parameter_values: the parameters for each circuit.
            args: additional arguments for the function
        """
        super().__init__(backend, job_id)
        self._fn = fn
        self._experiments = experiments
        self._parameter_values = parameter_values
        self._args = args
        self._future: futures.Future[Result] | None = None

    def submit(self) -> None:
        """Submit the job to the backend for execution.

        Raises:
            JobError: if trying to re-submit the job.
        """
        if self._future is not None:
            msg = "Job was already submitted!"
            raise JobError(msg)

        self._future = self._executor.submit(
            self._fn,
            self._job_id,
            self._experiments,
            self._parameter_values,
            **self._args,
        )

    @requires_submit
    def result(self, timeout: float | None = None) -> Result:
        # pylint: disable=arguments-differ
        """Get job result.

        The behavior is the same as the underlying
        concurrent Future objects,
        https://docs.python.org/3/library/concurrent.futures.html#future-objects.

        Args:
            timeout: number of seconds to wait for results.

        Returns:
            Result object
        Raises:
            concurrent.futures.TimeoutError: if timeout occurred.
            concurrent.futures.CancelledError: if job cancelled before completed.
        """
        assert self._future is not None
        return self._future.result(timeout=timeout)

    @requires_submit
    def cancel(self) -> bool:
        """Attempt to cancel the job."""
        assert self._future is not None
        return self._future.cancel()

    @requires_submit
    def status(self) -> JobStatus:
        """Gets the status of the job by querying the Python's future.

        Returns:
            JobStatus: The current JobStatus
        Raises:
            JobError: If the future is in unexpected state
            concurrent.futures.TimeoutError: if timeout occurred.
        """
        # The order is important here
        assert self._future is not None
        if self._future.running():
            return JobStatus.RUNNING
        if self._future.cancelled():
            return JobStatus.CANCELLED
        if self._future.done():
            return JobStatus.DONE if self._future.exception() is None else JobStatus.ERROR

        # Note: There is an undocumented Future state: PENDING, that seems to show up when
        # the job is enqueued, waiting for someone to pick it up. We need to deal with this
        # state but there's no public API for it, so we are assuming that if the job is not
        # in any of the previous states, is PENDING, ergo INITIALIZING for us.
        return JobStatus.INITIALIZING

    def backend(self) -> BackendV2 | None:
        """Return the instance of the backend used for this job."""
        return self._backend



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Qiskit backend for MQT DDSIM QASM simulators."""

from __future__ import annotations

import time
import uuid
from typing import TYPE_CHECKING, Any, cast

import numpy as np
from mqt.core import load
from qiskit import QuantumCircuit
from qiskit.providers import BackendV2, Options
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData
from qiskit.transpiler import Target

from . import __version__
from .experiment_header import DDSIMExperimentHeader
from .job import DDSIMJob
from .pyddsim import CircuitSimulator
from .target import DDSIMTargetBuilder

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from qiskit.circuit import Parameter
    from qiskit.circuit.parameterexpression import ParameterValueType

    Parameters = Mapping[Parameter, ParameterValueType] | Sequence[ParameterValueType]


class QasmSimulatorBackend(BackendV2):  # type: ignore[misc]
    """Qiskit backend for MQT DDSIM QASM simulators."""

    _SHOW_STATE_VECTOR = False
    _TARGET = Target(
        description="Target for the MQT DDSIM QASM simulator",
        num_qubits=128,
    )

    @staticmethod
    def _add_operations_to_target(target: Target) -> None:
        DDSIMTargetBuilder.add_0q_gates(target)
        DDSIMTargetBuilder.add_1q_gates(target)
        DDSIMTargetBuilder.add_2q_gates(target)
        DDSIMTargetBuilder.add_3q_gates(target)
        DDSIMTargetBuilder.add_multi_qubit_gates(target)
        DDSIMTargetBuilder.add_non_unitary_operations(target)
        DDSIMTargetBuilder.add_barrier(target)

    def _initialize_target(self) -> None:
        if len(self.target.operations) > 0:
            return
        self._add_operations_to_target(self.target)

    def __init__(
        self,
        name: str = "qasm_simulator",
        description: str = "MQT DDSIM QASM simulator",
    ) -> None:
        """Constructor for the MQT DDSIM QASM simulator backend."""
        super().__init__(name=name, description=description, backend_version=__version__)
        self._initialize_target()

    @classmethod
    def _default_options(cls) -> Options:
        return Options(
            shots=1024,
            memory=False,
            parameter_binds=None,
            seed_simulator=None,
            approximation_step_fidelity=1.0,
            approximation_steps=1,
            approximation_strategy="fidelity",
        )

    @property
    def target(self) -> Target:
        """Return the target of the backend."""
        return self._TARGET

    @property
    def max_circuits(self) -> int | None:
        """Return the maximum number of circuits that can be run in a single job."""
        return None

    @staticmethod
    def assign_parameters(
        quantum_circuits: Sequence[QuantumCircuit],
        parameter_values: Sequence[Parameters] | None,
    ) -> list[QuantumCircuit]:
        """Assign parameter values to the circuits.

        Args:
            quantum_circuits: The quantum circuits to assign parameters to.
            parameter_values: The parameter values to bind to the circuits.

        Returns:
            The bound circuits.

        Raises:
            ValueError: If the number of circuits does not match the number of provided parameter sets.
        """
        if not any(qc.parameters for qc in quantum_circuits) and not parameter_values:
            return list(quantum_circuits)

        if parameter_values is None:
            msg = "No parameter values provided although at least one parameterized circuit was supplied."
            raise ValueError(msg)

        if len(quantum_circuits) != len(parameter_values):
            msg = f"The number of circuits ({len(quantum_circuits)}) does not match the number of provided parameter sets ({len(parameter_values)})."
            raise ValueError(msg)

        bound_circuits = [
            qc.assign_parameters(parameters=values)
            for qc, values in zip(quantum_circuits, parameter_values, strict=False)
        ]

        # fix the circuit names
        for qcb, qc in zip(bound_circuits, quantum_circuits, strict=False):
            qcb.name = qc.name

        return bound_circuits

    def run(
        self,
        quantum_circuits: QuantumCircuit | Sequence[QuantumCircuit],
        parameter_values: Sequence[Parameters] | None = None,
        **options: Any,
    ) -> DDSIMJob:
        """Run a quantum circuit or list of quantum circuits on the DDSIM backend.

        Args:
            quantum_circuits: The quantum circuit(s) to run.
            parameter_values: The parameter values to bind to the circuits.
            options: Additional run options.

        Returns:
            The DDSIM job
        """
        if isinstance(quantum_circuits, QuantumCircuit):
            quantum_circuits = [quantum_circuits]

        job_id = str(uuid.uuid4())
        local_job = DDSIMJob(self, job_id, self._run_job, quantum_circuits, parameter_values, **options)
        local_job.submit()
        return local_job

    def _validate(self, quantum_circuits: Sequence[QuantumCircuit]) -> None:
        pass

    def _run_job(
        self,
        job_id: int,
        quantum_circuits: Sequence[QuantumCircuit],
        parameter_values: Sequence[Parameters] | None,
        **options: Any,
    ) -> Result:
        self._validate(quantum_circuits)
        start = time.time()

        bound_circuits = self.assign_parameters(quantum_circuits, parameter_values)
        result_list = [self._run_experiment(q_circ, **options) for q_circ in bound_circuits]

        end = time.time()

        return Result(
            backend_name=self.name,
            backend_version=self.backend_version,
            qobj_id=None,
            job_id=job_id,
            success=all(res.success for res in result_list),
            results=result_list,
            status="COMPLETED",
            time_taken=end - start,
        )

    def _run_experiment(self, qc: QuantumCircuit, **options: dict[str, Any]) -> ExperimentResult:
        start_time = time.time()
        approximation_step_fidelity = cast("float", options.get("approximation_step_fidelity", 1.0))
        approximation_steps = cast("int", options.get("approximation_steps", 1))
        approximation_strategy = str(options.get("approximation_strategy", "fidelity"))
        seed = cast("int", options.get("seed_simulator", -1))
        shots = cast("int", options.get("shots", 1024))

        circuit = load(qc)
        sim = CircuitSimulator(
            circuit,
            approximation_step_fidelity=approximation_step_fidelity,
            approximation_steps=approximation_steps,
            approximation_strategy=approximation_strategy,
            seed=seed,
        )
        counts = sim.simulate(shots=shots)
        end_time = time.time()

        data = ExperimentResultData(
            counts={hex(int(result, 2)): count for result, count in counts.items()},
            statevector=None if not self._SHOW_STATE_VECTOR else np.array(sim.get_constructed_dd().get_vector()),
            time_taken=end_time - start_time,
        )

        return ExperimentResult(
            shots=shots,
            success=True,
            status="DONE",
            seed=seed,
            data=data,
            metadata=qc.metadata,
            header=DDSIMExperimentHeader.from_quantum_circuit(qc).get_compatible_version(),
        )



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Nox sessions."""

from __future__ import annotations

import argparse
import contextlib
import os
import shutil
import sys
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING

import nox

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence

nox.needs_version = ">=2024.3.2"
nox.options.default_venv_backend = "uv"

nox.options.sessions = ["lint", "tests", "minimums"]

PYTHON_ALL_VERSIONS = ["3.10", "3.11", "3.12", "3.13", "3.14"]

if os.environ.get("CI", None):
    nox.options.error_on_missing_interpreters = True


@contextlib.contextmanager
def preserve_lockfile() -> Generator[None]:
    """Preserve the lockfile by moving it to a temporary directory."""
    with tempfile.TemporaryDirectory() as temp_dir_name:
        shutil.move("uv.lock", f"{temp_dir_name}/uv.lock")
        try:
            yield
        finally:
            shutil.move(f"{temp_dir_name}/uv.lock", "uv.lock")


@nox.session(reuse_venv=True)
def lint(session: nox.Session) -> None:
    """Run the linter."""
    if shutil.which("pre-commit") is None:
        session.install("pre-commit")

    session.run("pre-commit", "run", "--all-files", *session.posargs, external=True)


def _run_tests(
    session: nox.Session,
    *,
    install_args: Sequence[str] = (),
    extra_command: Sequence[str] = (),
    run_args: Sequence[str] = (),
) -> None:
    env = {"UV_PROJECT_ENVIRONMENT": session.virtualenv.location}
    if os.environ.get("CI", None) and sys.platform == "win32":
        env["SKBUILD_CMAKE_ARGS"] = "-T ClangCL"

    if shutil.which("cmake") is None and shutil.which("cmake3") is None:
        session.install("cmake")
    if shutil.which("ninja") is None:
        session.install("ninja")

    # install build and test dependencies on top of the existing environment
    session.run(
        "uv",
        "sync",
        "--inexact",
        "--only-group",
        "build",
        "--only-group",
        "test",
        *install_args,
        env=env,
    )
    if extra_command:
        session.run(*extra_command, env=env)
    if "--cov" in session.posargs:
        # try to use the lighter-weight `sys.monitoring` coverage core
        env["COVERAGE_CORE"] = "sysmon"
    session.run(
        "uv",
        "run",
        "--no-dev",  # do not auto-install dev dependencies
        "--no-build-isolation-package",
        "mqt-ddsim",  # build the project without isolation
        *install_args,
        "pytest",
        *run_args,
        *session.posargs,
        "--cov-config=pyproject.toml",
        env=env,
    )


@nox.session(reuse_venv=True, python=PYTHON_ALL_VERSIONS)
def tests(session: nox.Session) -> None:
    """Run the test suite."""
    _run_tests(session)


@nox.session(reuse_venv=True, venv_backend="uv", python=PYTHON_ALL_VERSIONS)
def minimums(session: nox.Session) -> None:
    """Test the minimum versions of dependencies."""
    with preserve_lockfile():
        _run_tests(
            session,
            install_args=["--resolution=lowest-direct"],
            run_args=["-Wdefault"],
        )
        env = {"UV_PROJECT_ENVIRONMENT": session.virtualenv.location}
        session.run("uv", "tree", "--frozen", env=env)


@nox.session(reuse_venv=True, venv_backend="uv", python=PYTHON_ALL_VERSIONS)
def qiskit(session: nox.Session) -> None:
    """Tests against the latest version of Qiskit."""
    with preserve_lockfile():
        _run_tests(
            session,
            extra_command=["uv", "pip", "install", "qiskit[qasm3-import] @ git+https://github.com/Qiskit/qiskit.git"],
        )
        env = {"UV_PROJECT_ENVIRONMENT": session.virtualenv.location}
        session.run("uv", "pip", "show", "qiskit", env=env)


@nox.session(reuse_venv=True)
def docs(session: nox.Session) -> None:
    """Build the docs. Use "--non-interactive" to avoid serving. Pass "-b linkcheck" to check links."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-b", dest="builder", default="html", help="Build target (default: html)")
    args, posargs = parser.parse_known_args(session.posargs)

    serve = args.builder == "html" and session.interactive
    if serve:
        session.install("sphinx-autobuild")

    env = {"UV_PROJECT_ENVIRONMENT": session.virtualenv.location}
    # install build and docs dependencies on top of the existing environment
    session.run(
        "uv",
        "sync",
        "--inexact",
        "--only-group",
        "build",
        "--only-group",
        "docs",
        env=env,
    )

    # build the C++ API docs using doxygen
    with session.chdir("docs"):
        if shutil.which("doxygen") is None:
            session.error("doxygen is required to build the C++ API docs")

        Path("_build/doxygen").mkdir(parents=True, exist_ok=True)
        session.run("doxygen", "Doxyfile", external=True)
        Path("api/cpp").mkdir(parents=True, exist_ok=True)
        session.run(
            "breathe-apidoc",
            "-o",
            "api/cpp",
            "-m",
            "-f",
            "-g",
            "file",
            "_build/doxygen/xml/",
            external=True,
        )

    shared_args = (
        "-n",  # nitpicky mode
        "-T",  # full tracebacks
        f"-b={args.builder}",
        "docs",
        f"docs/_build/{args.builder}",
        *posargs,
    )

    session.run(
        "uv",
        "run",
        "--no-dev",  # do not auto-install dev dependencies
        "--no-build-isolation-package",
        "mqt-ddsim",  # build the project without isolation
        "sphinx-autobuild" if serve else "sphinx-build",
        *shared_args,
        env=env,
    )



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Utilities for constructing a Qiskit experiment header for DDSIM backends."""

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Any

try:
    from qiskit.result.models import QobjExperimentHeader

    QISKIT_PRE_2_0 = True
except ImportError:
    QobjExperimentHeader = object
    QISKIT_PRE_2_0 = False

if TYPE_CHECKING:
    from qiskit import QuantumCircuit


@dataclass
class DDSIMExperimentHeader(QobjExperimentHeader):  # type: ignore[misc]
    """Header for DDSIM backends."""

    name: str
    n_qubits: int
    memory_slots: int
    global_phase: float
    creg_sizes: list[tuple[str, int]]
    clbit_labels: list[tuple[str, int]]
    qreg_sizes: list[tuple[str, int]]
    qubit_labels: list[tuple[str, int]]

    @classmethod
    def from_quantum_circuit(cls, quantum_circuit: QuantumCircuit) -> DDSIMExperimentHeader:
        """Create a DDSIM experiment header from a QuantumCircuit."""
        return cls(
            name=quantum_circuit.name,
            n_qubits=quantum_circuit.num_qubits,
            memory_slots=quantum_circuit.num_clbits,
            global_phase=quantum_circuit.global_phase,
            creg_sizes=[(creg.name, creg.size) for creg in quantum_circuit.cregs],
            clbit_labels=[(creg.name, j) for creg in quantum_circuit.cregs for j in range(creg.size)],
            qreg_sizes=[(qreg.name, qreg.size) for qreg in quantum_circuit.qregs],
            qubit_labels=[(qreg.name, j) for qreg in quantum_circuit.qregs for j in range(qreg.size)],
        )

    def get_compatible_version(self) -> QobjExperimentHeader | dict[str, Any]:
        """Return a compatible header.

        For Qiskit < 2.0, return a QobjExperimentHeader. For Qiskit >= 2.0, return a dict.
        """
        if QISKIT_PRE_2_0:
            return self

        return asdict(self)



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""MQT DDSIM Python Package."""

from __future__ import annotations

import sys

# under Windows, make sure to add the appropriate DLL directory to the PATH
if sys.platform == "win32":

    def _dll_patch() -> None:
        """Add the DLL directory to the PATH."""
        import os
        import sysconfig
        from pathlib import Path

        site_packages = Path(sysconfig.get_paths()["purelib"])
        bin_dir = site_packages / "mqt" / "core" / "bin"
        os.add_dll_directory(str(bin_dir))

    _dll_patch()
    del _dll_patch

from ._version import version as __version__
from .provider import DDSIMProvider
from .pyddsim import (
    CircuitSimulator,
    DeterministicNoiseSimulator,
    HybridSimulator,
    HybridSimulatorMode,
    PathSimulator,
    PathSimulatorConfiguration,
    PathSimulatorMode,
    StochasticNoiseSimulator,
    UnitarySimulator,
    UnitarySimulatorMode,
)

__all__ = [
    "CircuitSimulator",
    "DDSIMProvider",
    "DeterministicNoiseSimulator",
    "HybridSimulator",
    "HybridSimulatorMode",
    "PathSimulator",
    "PathSimulatorConfiguration",
    "PathSimulatorMode",
    "StochasticNoiseSimulator",
    "UnitarySimulator",
    "UnitarySimulatorMode",
    "__version__",
]
